package com.fjhx.flow.service.flow.impl;

import cn.hutool.core.util.ObjectUtil;
import cn.hutool.core.util.StrUtil;
import com.alibaba.fastjson.JSONObject;
import com.baomidou.dynamic.datasource.annotation.DSTransactional;
import com.baomidou.mybatisplus.core.toolkit.IdWorker;
import com.fjhx.flow.core.FlowBean;
import com.fjhx.flow.core.FlowDelegate;
import com.fjhx.flow.core.FlowThreadLocalUtil;
import com.fjhx.flow.entity.flow.dto.FlowResult;
import com.fjhx.flow.entity.flow.dto.InitiateDto;
import com.fjhx.flow.entity.flow.dto.JumpDto;
import com.fjhx.flow.entity.flow.po.FlowDefinition;
import com.fjhx.flow.entity.flow.po.FlowDefinitionNode;
import com.fjhx.flow.entity.flow.po.FlowExample;
import com.fjhx.flow.entity.flow.po.FlowExampleDetail;
import com.fjhx.flow.enums.FlowStatusEnum;
import com.fjhx.flow.enums.HandleObjectTypeEnum;
import com.fjhx.flow.enums.HandleTypeEnum;
import com.fjhx.flow.enums.NodeTypeEnum;
import com.fjhx.flow.service.flow.*;
import com.googlecode.aviator.AviatorEvaluator;
import com.googlecode.aviator.Expression;
import com.ruoyi.common.constant.StatusConstant;
import com.ruoyi.common.core.domain.entity.SysDept;
import com.ruoyi.common.core.domain.entity.SysUser;
import com.ruoyi.common.exception.ServiceException;
import com.ruoyi.system.service.ISysDeptService;
import com.ruoyi.system.service.ISysUserService;
import lombok.extern.slf4j.Slf4j;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Service;

import java.lang.reflect.Method;
import java.util.ArrayList;
import java.util.List;
import java.util.Map;
import java.util.function.Function;
import java.util.regex.Matcher;
import java.util.regex.Pattern;
import java.util.stream.Collectors;

@Slf4j
@Service
public class FlowProcessServiceImpl implements FlowProcessService {

    @Autowired
    private FlowDefinitionService flowDefinitionService;

    @Autowired
    private FlowDefinitionNodeService flowDefinitionNodeService;

    @Autowired
    private FlowExampleService flowExampleService;

    @Autowired
    private FlowExampleDetailService flowExampleDetailService;

    @Autowired
    private ISysDeptService sysDeptService;

    @Autowired
    private ISysUserService sysUserService;

    @DSTransactional
    @Override
    public FlowResult initiate(InitiateDto dto) {

        // 获取流程委托对象
        FlowDelegate flowDelegate = FlowBean.getBean(dto.getFlowKey());

        // 查找可用流程
        FlowDefinition flowDefinition = flowDefinitionService.getOne(q -> q
                .eq(FlowDefinition::getFlowKey, dto.getFlowKey())
                .eq(FlowDefinition::getCurrentVersion, StatusConstant.YES));

        if (flowDefinition == null) {
            throw new ServiceException("流程未配置，请联系管理员");
        }

        // 流程节点列表
        List<FlowDefinitionNode> flowDefinitionNodeList = flowDefinitionNodeService.list(q -> q
                .eq(FlowDefinitionNode::getFlowDefinitionId, flowDefinition.getId()));

        // 寻找开始节点
        FlowDefinitionNode startNode = flowDefinitionNodeList.stream()
                .filter(item -> NodeTypeEnum.START.getKey().equals(item.getNodeType()))
                .findFirst().orElse(null);

        if (startNode == null) {
            throw new ServiceException("没有找到开始节点");
        }

        // 生成流程id
        final long flowId = IdWorker.getId();

        // 初始化模板参数
        Map<String, Object> templateMap = flowDelegate.initTemplateMap(dto.getData(), dto.getData());
        FlowThreadLocalUtil.setTemplateData(templateMap);
        FlowThreadLocalUtil.setStartData(dto.getData());
        FlowThreadLocalUtil.setCurrentData(dto.getData());
        FlowThreadLocalUtil.setFlowId(flowId);

        // 寻找下一节点
        FlowDefinitionNode nextUserNode = getNextUserNode(startNode, flowDefinitionNodeList);
        NodeTypeEnum nextNodeType = NodeTypeEnum.getEnum(nextUserNode.getNodeType());

        // 流程实例
        FlowExample flowExample = new FlowExample();

        // 如果下一节点不为结束节点，赋值下一个节点处理用户id
        if (!NodeTypeEnum.END.equals(nextNodeType)) {

            FlowResult handleUser = getHandleUser(nextUserNode, dto.getHandleUserId());
            // 如果下一节点处理用户只有1人，赋值用户id
            if (handleUser.getSuccess()) {
                flowExample.setHandleUserId(handleUser.getUserId());
                FlowThreadLocalUtil.setNextHandleUserId(handleUser.getUserId());
            }
            // 如果下一节点处理用户有多人，则返回用户列表让用户选择下一节点处理人id
            else {
                return handleUser;
            }
            // 流程进行中
            flowExample.setStatus(FlowStatusEnum.IN_PROGRESS.getKey());
        } else {
            // 流程已通过
            flowExample.setStatus(FlowStatusEnum.PASS.getKey());
        }

        // 执行开始流程方法
        Long businessId;
        try {
            businessId = flowDelegate.start(flowId, dto.getData());
        } catch (Exception e) {
            log.error("开始节点方法异常", e);
            throw new ServiceException("开始节点方法异常：" + e.getMessage());
        }

        List<FlowExampleDetail> flowExampleDetailList = new ArrayList<>();

        // 开始节点明细
        FlowExampleDetail startExampleDetail = new FlowExampleDetail();
        startExampleDetail.setFlowExampleId(flowId);
        startExampleDetail.setFlowDefinitionNodeId(startNode.getId());
        startExampleDetail.setFlowDefinitionNodeType(startNode.getNodeType());
        startExampleDetail.setHandleType(HandleTypeEnum.SKIP_TO_NEXT.getKey());
        startExampleDetail.setHandleRemark(dto.getRemark());
        startExampleDetail.setSubmitData(JSONObject.toJSONString(dto));
        flowExampleDetailList.add(startExampleDetail);

        if (NodeTypeEnum.END.equals(nextNodeType)) {

            // 赋值业务id到租户线程
            FlowThreadLocalUtil.setBusinessId(businessId);

            // 执行结束方法
            invokeEndMethod(nextUserNode, flowDelegate);

            // 结束流程明细
            FlowExampleDetail endExampleDetail = new FlowExampleDetail();
            endExampleDetail.setFlowExampleId(flowId);
            endExampleDetail.setFlowDefinitionNodeId(nextUserNode.getId());
            endExampleDetail.setFlowDefinitionNodeType(nextUserNode.getNodeType());
            endExampleDetail.setHandleType(HandleTypeEnum.SKIP_TO_NEXT.getKey());
            flowExampleDetailList.add(endExampleDetail);
        }

        flowExample.setId(flowId);
        flowExample.setTitle(templateParse(flowDefinition.getTitleTemplate(), templateMap));
        flowExample.setFlowKey(dto.getFlowKey());
        flowExample.setDefinitionId(flowDefinition.getId());
        flowExample.setDefinitionNodeId(nextUserNode.getId());
        flowExample.setBusinessId(businessId);
        flowExample.setStartData(dto.getData().toJSONString());
        flowExample.setDefinitionVersion(flowDefinition.getVersionNumber());

        flowExampleService.save(flowExample);
        flowExampleDetailService.saveBatch(flowExampleDetailList);

        FlowResult flowResult = new FlowResult();
        flowResult.setSuccess(true);
        return flowResult;
    }

    @DSTransactional
    @Override
    public FlowResult jump(JumpDto dto) {
        Long flowId = dto.getFlowId();

        FlowExample flowExample = flowExampleService.getById(flowId);

        if (flowExample == null) {
            throw new ServiceException("没有找到流程");
        }

        // if (flowExample.getStatus() > 1) {
        //     throw new ServiceException("流程已结束");
        // }

        if (ObjectUtil.notEqual(flowExample.getVersion(), dto.getVersion())) {
            throw new ServiceException("流程已被处理");
        }

        // 获取流程委托对象
        FlowDelegate flowDelegate = FlowBean.getBean(flowExample.getFlowKey());

        // 流程节点列表
        List<FlowDefinitionNode> flowDefinitionNodeList = flowDefinitionNodeService.list(q -> q
                .eq(FlowDefinitionNode::getFlowDefinitionId, flowExample.getDefinitionId()));

        // 获取当前审批节点
        FlowDefinitionNode currentNode = flowDefinitionNodeList.stream()
                .filter(item -> item.getId().equals(flowExample.getDefinitionNodeId()))
                .findFirst().orElse(null);

        if (currentNode == null) {
            throw new ServiceException("没有找到当前审批节点");
        }

        // 查找跳转类型
        HandleTypeEnum handleTypeEnum = HandleTypeEnum.getEnum(dto.getHandleType());

        // 发起流程参数
        JSONObject startData = JSONObject.parseObject(flowExample.getStartData());
        Map<String, Object> templateMap = flowDelegate.initTemplateMap(dto.getData(), startData);
        FlowThreadLocalUtil.setTemplateData(templateMap);
        FlowThreadLocalUtil.setStartData(startData);
        FlowThreadLocalUtil.setCurrentData(dto.getData());
        FlowThreadLocalUtil.setFlowId(flowId);
        FlowThreadLocalUtil.setBusinessId(flowExample.getBusinessId());
        FlowThreadLocalUtil.setHandleTypeEnum(handleTypeEnum);

        // 跳转节点
        FlowDefinitionNode nextUserNode = null;
        switch (handleTypeEnum) {

            // 跳转下一节点
            case SKIP_TO_NEXT:

                // 寻找下一节点
                nextUserNode = getNextUserNode(currentNode, flowDefinitionNodeList);

                // 赋值流程所在id
                flowExample.setDefinitionNodeId(nextUserNode.getId());

                // 如果下一节点为结束节点
                if (NodeTypeEnum.END.equals(NodeTypeEnum.getEnum(nextUserNode.getNodeType()))) {
                    // 流程已通过
                    flowExample.setStatus(FlowStatusEnum.PASS.getKey());
                }
                // 如果下一节点不为结束节点
                else {
                    FlowResult handleUser = getHandleUser(nextUserNode, dto.getHandleUserId());

                    // 如果下一节点处理用户只有1人，赋值用户id
                    if (handleUser.getSuccess()) {
                        flowExample.setHandleUserId(handleUser.getUserId());
                        FlowThreadLocalUtil.setNextHandleUserId(handleUser.getUserId());
                    }
                    // 如果下一节点处理用户有多人，则返回用户列表让用户选择下一节点处理人id
                    else {
                        return handleUser;
                    }

                    // 流程进行中
                    flowExample.setStatus(FlowStatusEnum.IN_PROGRESS.getKey());
                }
                break;

            // 结束流程
            case REJECT:
                flowExample.setStatus(FlowStatusEnum.REJECT.getKey());
                break;

            // 返回上一节点
            case RETURN_TO_PREVIOUS:

                // 查找上一个节点
                FlowDefinitionNode lastOneUserNode = getLastOneUserNode(currentNode, flowDefinitionNodeList);

                // 如果流程回退到开始节点
                if (NodeTypeEnum.START.getKey().equals(lastOneUserNode.getNodeType())) {
                    // 流程未发起
                    flowExample.setStatus(FlowStatusEnum.READY_START.getKey());
                } else {
                    FlowResult handleUser = getHandleUser(lastOneUserNode, dto.getHandleUserId());

                    // 如果上一节点处理用户只有1人，赋值用户id
                    if (handleUser.getSuccess()) {
                        flowExample.setHandleUserId(handleUser.getUserId());
                        FlowThreadLocalUtil.setNextHandleUserId(handleUser.getUserId());
                    }
                    // 如果下一节点处理用户有多人，则返回用户列表让用户选择下一节点处理人id
                    else {
                        return handleUser;
                    }

                    // 流程进行中
                    flowExample.setStatus(FlowStatusEnum.IN_PROGRESS.getKey());
                }
                break;

            // 退回到发起人
            case RETURN_TO_SUBMITTER:

                // 获取第一个用户操作节点
                FlowExampleDetail exampleDetail = flowExampleDetailService.getOne(q -> q
                        .eq(FlowExampleDetail::getFlowExampleId, flowId));

                flowExample.setHandleUserId(exampleDetail.getCreateUser());
                flowExample.setDefinitionNodeId(exampleDetail.getFlowDefinitionNodeId());
                flowExample.setStatus(FlowStatusEnum.READY_START.getKey());
                FlowThreadLocalUtil.setNextHandleUserId(exampleDetail.getCreateUser());
                break;

            default:
                throw new ServiceException("未知节点处理类型");

        }

        // 执行当前节点方法
        String handlingMethod = currentNode.getHandlingMethod();
        if (StrUtil.isNotBlank(handlingMethod)) {
            try {
                invokeMethod(flowDelegate, handlingMethod);
            } catch (Exception e) {
                log.error("跳转节点方法异常", e);
                throw new ServiceException("跳转节点方法异常：" + e.getMessage());
            }
        }

        // 节点处理明细
        List<FlowExampleDetail> flowExampleDetailList = new ArrayList<>();
        FlowExampleDetail nodeExampleDetail = new FlowExampleDetail();
        nodeExampleDetail.setFlowExampleId(flowId);
        nodeExampleDetail.setFlowDefinitionNodeId(currentNode.getId());
        nodeExampleDetail.setFlowDefinitionNodeType(currentNode.getNodeType());
        nodeExampleDetail.setHandleType(dto.getHandleType());
        nodeExampleDetail.setHandleRemark(dto.getRemark());
        nodeExampleDetail.setSubmitData(JSONObject.toJSONString(dto));
        flowExampleDetailList.add(nodeExampleDetail);

        // 如果流程已结束
        if (FlowStatusEnum.PASS.getKey().equals(flowExample.getStatus())) {

            // 执行结束方法
            invokeEndMethod(nextUserNode, flowDelegate);

            // 结束流程明细
            FlowExampleDetail endExampleDetail = new FlowExampleDetail();
            endExampleDetail.setFlowExampleId(flowId);
            endExampleDetail.setFlowDefinitionNodeId(nextUserNode.getId());
            endExampleDetail.setFlowDefinitionNodeType(nextUserNode.getNodeType());
            endExampleDetail.setHandleType(HandleTypeEnum.SKIP_TO_NEXT.getKey());
            flowExampleDetailList.add(endExampleDetail);

        }

        flowExample.setVersion(dto.getVersion());
        boolean updateFlag = flowExampleService.updateById(flowExample);
        if (!updateFlag) {
            throw new ServiceException("流程已被处理");
        }
        flowExampleDetailService.saveBatch(flowExampleDetailList);

        FlowResult flowResult = new FlowResult();
        flowResult.setSuccess(true);
        return flowResult;
    }

    @Override
    public JSONObject getStartData(Long flowId) {
        if (flowId == null) {
            throw new ServiceException("流程实例id不能为空");
        }
        FlowExample flowExample = flowExampleService.getById(flowId);
        if (flowExample == null) {
            throw new ServiceException("没有找到流程实例");
        }
        return JSONObject.parseObject(flowExample.getStartData());
    }

    /**
     * 执行节点方法
     *
     * @param flowDelegate   流程代理对象
     * @param handlingMethod 流程执行方法名
     */
    private void invokeMethod(FlowDelegate flowDelegate, String handlingMethod) throws Exception {
        Class<? extends FlowDelegate> flowDelegateCls = flowDelegate.getClass();
        Method method = flowDelegateCls.getMethod(handlingMethod.trim());
        method.invoke(flowDelegate);
    }

    /**
     * 执行结束节点方法
     *
     * @param endNode      结束节点
     * @param flowDelegate java委托对象
     */
    private void invokeEndMethod(FlowDefinitionNode endNode, FlowDelegate flowDelegate) {
        try {
            // 执行结束方法
            String handlingMethod = endNode.getHandlingMethod();
            if (StrUtil.isNotBlank(handlingMethod)) {
                // 执行指定方法
                invokeMethod(flowDelegate, handlingMethod);
            } else {
                // 执行默认方法
                flowDelegate.end(
                        FlowThreadLocalUtil.getFlowId(),
                        FlowThreadLocalUtil.getBusinessId(),
                        FlowThreadLocalUtil.getCurrentData());
            }
        } catch (Exception e) {
            log.error("结束节点方法异常", e);
            throw new ServiceException("结束节点方法异常：" + e.getMessage());
        }
    }

    /**
     * 查找上一个节点
     *
     * @param currentNode            当前节点
     * @param flowDefinitionNodeList 流程节点列表
     * @return 用户节点
     */
    private FlowDefinitionNode getLastOneUserNode(FlowDefinitionNode currentNode, List<FlowDefinitionNode> flowDefinitionNodeList) {
        // 节点map
        Map<Long, FlowDefinitionNode> nodeMap = flowDefinitionNodeList.stream()
                .collect(Collectors.toMap(FlowDefinitionNode::getId, Function.identity()));

        // 查找上个节点
        FlowDefinitionNode lastNode = nodeMap.get(currentNode.getParentId());

        // 如果为分支
        if (NodeTypeEnum.BRANCH.equals(NodeTypeEnum.getEnum(lastNode.getNodeType()))) {
            return getLastOneUserNode(lastNode, flowDefinitionNodeList);
        }

        return lastNode;
    }

    /**
     * 查找下一个节点
     *
     * @param currentNode            当前节点
     * @param flowDefinitionNodeList 流程节点列表
     * @return 用户节点
     */
    private FlowDefinitionNode getNextUserNode(FlowDefinitionNode currentNode, List<FlowDefinitionNode> flowDefinitionNodeList) {
        // 父级节点map
        Map<Long, List<FlowDefinitionNode>> parentNodeMap =
                flowDefinitionNodeList.stream().collect(Collectors.groupingBy(FlowDefinitionNode::getParentId));
        // 查找下个节点
        List<FlowDefinitionNode> nextNodeList = parentNodeMap.get(currentNode.getId());
        FlowDefinitionNode nextNode = null;
        // 如果为分支
        if (NodeTypeEnum.BRANCH.equals(NodeTypeEnum.getEnum(currentNode.getNodeType()))) {
            // 循环分支
            for (FlowDefinitionNode flowDefinitionNode : nextNodeList) {
                // 获取跳转条件
                String jumpCondition = flowDefinitionNode.getJumpCondition();
                // 符合条件
                if (expressionResult(FlowThreadLocalUtil.getTemplateData(), jumpCondition)) {
                    nextNode = flowDefinitionNode;
                    break;
                }
                // 跳转条件为空表示默认跳转
                if (StrUtil.isBlank(jumpCondition)) {
                    nextNode = flowDefinitionNode;
                }
            }
            if (nextNode == null) {
                throw new ServiceException("未识别到分支跳转节点");
            }
        }
        // 如果不为分支
        else {
            if (nextNodeList == null || nextNodeList.size() != 1) {
                throw new ServiceException("流程定义错误");
            }
            // 查找下一个节点方法
            nextNode = nextNodeList.get(0);
        }
        // 如果下一节点为分支，递归找下一节点
        if (NodeTypeEnum.BRANCH.equals(NodeTypeEnum.getEnum(nextNode.getNodeType()))) {
            return getNextUserNode(nextNode, flowDefinitionNodeList);
        }
        return nextNode;
    }

    /**
     * 获取处理用户
     *
     * @param node 处理节点
     * @return 处理用户
     */
    private FlowResult getHandleUser(FlowDefinitionNode node, Long handleUserId) {
        FlowResult flowResult = new FlowResult();

        if (handleUserId != null) {
            flowResult.setSuccess(true);
            flowResult.setUserId(handleUserId);
            return flowResult;
        }

        Integer handleObjectType = node.getHandleObjectType();
        Long handleObjectId = node.getHandleObjectId();

        HandleObjectTypeEnum handleObjectTypeEnum = HandleObjectTypeEnum.getEnum(handleObjectType);

        switch (handleObjectTypeEnum) {

            case USER:
                flowResult.setSuccess(true);
                flowResult.setUserId(handleObjectId);
                return flowResult;

            case DETP_LEADER:
            case DEPT_DIRECTOR:
                SysDept sysDept = sysDeptService.getById(handleObjectId);
                if (sysDept == null) {
                    throw new ServiceException("部门为空");
                }
                // 部门负责人
                if (HandleObjectTypeEnum.DETP_LEADER.equals(handleObjectTypeEnum)) {
                    Long leaderId = sysDept.getLeaderId();
                    if (leaderId == null) {
                        throw new ServiceException("部门负责人为空");
                    }
                    flowResult.setUserId(leaderId);
                }
                // 部门总监
                else {
                    Long directorId = sysDept.getDirectorId();
                    if (directorId == null) {
                        throw new ServiceException("部门总监为空");
                    }
                    flowResult.setUserId(directorId);
                }
                flowResult.setSuccess(true);
                return flowResult;

            case POST:
                List<SysUser> postUserList = sysUserService.getListByPostId(handleObjectId);
                if (postUserList.size() == 0) {
                    throw new ServiceException("岗位无用户");
                }
                if (postUserList.size() == 1) {
                    flowResult.setSuccess(true);
                    flowResult.setUserId(postUserList.get(0).getUserId());
                    return flowResult;
                }
                flowResult.setSuccess(false);
                flowResult.setUserList(postUserList);
                return flowResult;

            case ROLE:
                List<SysUser> roleUserList = sysUserService.getListByRoleId(handleObjectId);
                if (roleUserList.size() == 0) {
                    throw new ServiceException("角色无用户");
                }
                if (roleUserList.size() == 1) {
                    flowResult.setSuccess(true);
                    flowResult.setUserId(roleUserList.get(0).getUserId());
                    return flowResult;
                }
                flowResult.setSuccess(false);
                flowResult.setUserList(roleUserList);
                return flowResult;

            default:
                throw new ServiceException("未知用户处理类型：" + handleObjectType);
        }

    }

    /**
     * 校验el表达式
     */
    public boolean expressionResult(Map<String, Object> map, String expression) {
        try {
            Expression exp = AviatorEvaluator.compile(expression);
            Object execute = exp.execute(map);
            return Boolean.parseBoolean(String.valueOf(execute));
        } catch (Exception e) {
            log.error("el表达式校验错误", e);
            throw new ServiceException("el表达式校验错误：" + e.getMessage());
        }
    }

    // public static void main(String[] args) {
    //     HashMap<String, Object> map = new HashMap<>();
    //
    //     HashMap<Object, Object> dd = new HashMap<>();
    //     dd.put("A", "a1");
    //     dd.put("B", "b1");
    //     dd.put("C", "c1");
    //
    //
    //     map.put("aa", "aa");
    //     map.put("bb", "bb");
    //     map.put("cc", Arrays.asList("A", "B", "C", "D"));
    //     map.put("dd", dd);
    //
    //     System.out.println(expressionResult(map,
    //             "aa == 'aa' && bb == 'bb' && cc[0] == 'A' && dd.A == 'a1' "
    //     ));
    //
    // }

    // public static void main(String[] args) {
    //
    //     System.out.println(
    //             templateParse(
    //                     "测试1：${Integer.parseInt(a)+1};  测试2：${a+1};  测试3：${b.tt};  测试4：${b.vv};  测试5：${b.vv[0]};",
    //                     JSONObject.parseObject("{a: \"1\", b: {tt: \"t\", vv: [\"aa\", \"bb\", \"cc\"] }}"))
    //     );
    //
    // }

    /**
     * 模板替换
     *
     * <p>
     * templateStr: ${Integer.parseInt(a)+1};  ${a+1};  ${b.tt};  ${b.vv};  ${b.vv[0]};
     * map:         {a: "1", b: {tt: "t", vv: ["aa", "bb", "cc"] }}
     * return:      2;  11;  t;  ["aa","bb","cc"];  aa;
     *
     * @param templateStr 模板字符串
     * @param map         替换参数
     * @return 替换后的字符串
     */
    private String templateParse(String templateStr, Map<String, Object> map) {
        Pattern pattern = Pattern.compile("\\$\\{(.*?)}");
        Matcher matcher = pattern.matcher(templateStr);
        StringBuffer sr = new StringBuffer();
        while (matcher.find()) {
            String group = matcher.group().replace("$", "");
            Object execute = AviatorEvaluator.compile(group).execute(map);
            if (execute != null) {
                matcher.appendReplacement(sr, execute.toString());
            }
        }
        matcher.appendTail(sr);
        return sr.toString();
    }

}
